import asyncio
import enum
import os, torch
import json
import queue
import threading
import uuid
from typing import Any, List, Tuple
from aiges.core.types import *

try:
    from aiges_embed import ResponseData, Response, DataListNode, DataListCls, SessionCreateResponse, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListNode, DataListCls, SessionCreateResponse, callback

from aiges.sdk import WrapperBase
from aiges.utils.log import log, getFileLogger
from belt_nlp.bert_classifier_with_pooling import BertClassifierWithPooling
from belt_nlp.bert import TokenizedDataset
from transformers import BertTokenizer
import logging
import threading


DataNone = -1
DataBegin = 0  # 首数据
DataContinue = 1  # 中间数据
DataEnd = 2  # 尾数据

class RequestMode(enum.Enum):
    ONCE = 'once'
    ONCE_ASYNC = 'once_async'
    STREAM = 'stream'
    STOP = "stop"

class RequestInfo:
    def __init__(self, sid: str, params: dict, user_tag: str="") -> None:
        self.handle = str(uuid.uuid4())
        self.sid = sid
        self.params = params
        self.user_tag = user_tag
        self.request = []      

class PromptInferenceInfo:
    def __init__(self, wrapper,
                 thread_id: str,
                 mode: RequestMode,
                 prompt: str,
                 requestInfo: RequestInfo,
                 result_q: queue.Queue = None):
        self.wrapper = wrapper
        self.thread_id = thread_id
        self.mode = mode
        self.prompt = prompt
        self.requestInfo = requestInfo
        self.result_q = result_q
        self.request_id = str(uuid.uuid4().hex)

elapsed_times = 0
import functools
import time
def time_cost(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        global elapsed_times
        elapsed_time = end_time - start_time
        elapsed_times += elapsed_time
        print(f"函数 {func.__name__} 的耗时为: {elapsed_time} 秒")
        return result
    return wrapper

def get_elapsed_time():
    return elapsed_times

def resp_content(status, index, text):
    resd = ResponseData()
    resd.key = "content"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(text).encode('utf-8'))
    return resd

def resp_usage(status, prompt_token, completion_token):
    resp_json = {"completion_tokens": completion_token, "prompt_tokens": prompt_token, "question_tokens": 4,
                 "total_tokens": prompt_token + completion_token}
    resd = ResponseData()
    resd.key = "usage"
    resd.setDataType(DataText)
    resd.status = status
    resd.setData(json.dumps(resp_json).encode('utf-8'))
    return resd

def get_payload_messages(reqData: DataListCls):
    messagesStr = reqData.get('messages').data.decode('utf-8')
    # 校验格式
    try:
        messages = json.loads(messagesStr)
        message = ""
        for item in messages['messages']:
            if item['role'] == "user":
                message = item['content']
    except Exception as e:
        import traceback
        traceback.print_exc()
        return None
    # self.filelogger.info(f'message: {message}')
    batch = []
    batch.append(message)
    return batch
    
class ThreadPool:
    def __init__(self, num_threads, wrapper):
        self.num_threads = num_threads
        self.threads = {}
        self.task_queues = {}
        self.lock = threading.Lock()
        self.wrapper = wrapper
        self.max_batch_size = 1
        self.filelogger = getFileLogger(level=logging.INFO)

        for i in range(num_threads):
            process_id = os.getpid
            thread_id = f"process-{str(process_id)}-thread-{str(i)}"
            task_queue = queue.Queue()

            self.threads[thread_id] = threading.Thread(target=self.worker, args=(thread_id, task_queue, self.max_batch_size))
            self.task_queues[thread_id] = task_queue

        for thread in self.threads.values():
            thread.start()

    # async def bert_loop(self, task_queue):
    #     self.filelogger.info(f"bert_loop started")
    #     while True:
    #         if not task_queue.empty():
    #             try:
    #                 task: PromptInferenceInfo = task_queue.get_nowait()
    #                 if task is None:
    #                     break
    #                 if task.mode == RequestMode.STOP:
    #                     await task.wrapper.infer(task)
    #                     continue
    #                 asyncio.create_task(task.wrapper.infer(task))
    #                 await asyncio.sleep(0)
    #             except Exception as e:
    #                 self.filelogger.error(f"Error processing task: {e}")
    #         else:
    #             await asyncio.sleep(0.01)
    #     self.filelogger.info("bert_loop end")    

    async def bert_loop(self, task_queue, max_batch_size):
        self.filelogger.info(f"bert_loop started")
        tasks = []
        while True:
            await asyncio.sleep(0.01)

            while not task_queue.empty() and len(tasks) < max_batch_size:
                task: PromptInferenceInfo = task_queue.get_nowait()
                if task is None:
                    break
                tasks.append(task)

            if tasks:
                self.filelogger.info(f"Processing {len(tasks)} tasks.")
                asyncio.create_task(task.wrapper.infer(tasks))
                tasks = [] 

    def worker(self, thread_id, task_queue, max_batch_size):
        # 创建事件循环
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.bert_loop(task_queue, max_batch_size))

    def alloc_min_thread(self) -> str:
        with self.lock:
            min_thread_id = min(self.threads, key=lambda thread_id: self.task_queues[thread_id].qsize())
        return min_thread_id
    
    def put_task(self, thread_id, task):
        with self.lock:
            self.task_queues[thread_id].put(task)

    def wait_completion(self):
        for task_queue in self.task_queues.values():
            task_queue.put(None)

        for thread in self.threads.values():
            thread.join()

class Wrapper(WrapperBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filelogger = getFileLogger(level=logging.INFO)
        self.patch_id = {}
        self.lock = threading.Lock()
        self.request_map = {}
        self.request_map_lock = threading.Lock()
        self.thread_pool_size = 1
        self.thread_pool = None
        
    def wrapperInit(self, config: dict) -> int:
        self.filelogger.info("Initializing ...")
        self.pretrained_model = os.environ.get("PRETRAINED_MODEL_NAME")
        self.base_model_path = os.environ.get("FULL_MODEL_PATH")
        self.tokenizer = BertTokenizer.from_pretrained(self.base_model_path, trust_remote_code=True)
        self.model = BertClassifierWithPooling.load(self.base_model_path)
        self.thread_pool = ThreadPool(num_threads=self.thread_pool_size, wrapper=self)
        id2label_file = os.path.join(self.base_model_path, 'id2label.json')
        with open(id2label_file, 'r') as file:
            tag_data = json.load(file)
        self.tag_data = tag_data
        self.filelogger.info(f'wrapper init success, create thread:{self.thread_pool_size}')
        return 0
    
    def wrapperFini(self):
        self.thread_pool.wait_completion()
        return 0
    
    def wrapperCreate(self, params: dict, sid: str, persId: int=0, usrTag: str="") -> SessionCreateResponse:
        patch_id = str(params.get("patch_id", "0"))
        self.filelogger.info(f'start wrapperCreate {params}')
        if len(patch_id) == 0:
            patch_id = "0"
        requestInfo = RequestInfo(sid, params, usrTag)
        self.request_map_lock.acquire()
        self.request_map[requestInfo.handle] = requestInfo
        self.request_map_lock.release()

        s = SessionCreateResponse()
        s.handle = requestInfo.handle
        s.error_code = 0
        self.filelogger.debug(f'success wrapperCreate {patch_id}, handle {requestInfo.handle}')
        return s

    def wrapperWrite(self, handle: str, req: DataListCls) -> str:
        self.filelogger.info(f'start wrapperWrite handle {handle}')
        prompt = get_payload_messages(req)
        if prompt is None:
            self.filelogger.error("parse prompt failed")
            return -1
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error(f"can't get this handle: {handle}")
            self.request_map_lock.release()
            return -1
        self.request_map_lock.release()
        thread_id = self.thread_pool.alloc_min_thread()
        inferenceInfo = PromptInferenceInfo(self, thread_id, RequestMode.STREAM, prompt, requestInfo)
        self.thread_pool.put_task(thread_id, inferenceInfo)
        self.request_map_lock.acquire()
        self.request_map[handle].request.append(inferenceInfo.request_id)
        self.request_map_lock.release()
        self.filelogger.debug(
            f'success wrapperWrite handle:{handle}, thread_id:{thread_id}, request_id:{inferenceInfo.request_id}')
        return 0

    def wrapperDestroy(self, handle: str):
        self.filelogger.debug(f'start wrapperDestroy {handle}')
        self.request_map_lock.acquire()
        requestInfo = self.request_map[handle]
        if requestInfo == None:
            self.filelogger.error(f"can't get this handle: {handle}")
            self.request_map_lock.release()
            return -1
        self.request_map_lock.release()
        self.filelogger.debug(f'success wrapperDestroy {handle}')
        return 0

    def wrapperError(self):
        pass

    def wrapperOnceExec(self):
        pass

    # async def infer(self, inferenceInfo: PromptInferenceInfo):
    #     requestInfo = inferenceInfo.requestInfo
    #     if inferenceInfo.mode == RequestMode.STOP:
    #         for req in requestInfo.request:
    #             inferenceInfo.result_q.put(req)
    #         return
    #     params = requestInfo.params
    #     patch_id = str(params.get("patch_id", 0))
    #     if len(patch_id) == 0:
    #         patch_id = '0'
    #     request_id = inferenceInfo.request_id
    #     sid = requestInfo.sid
    #     user_tag = requestInfo.user_tag
    #     prompt = inferenceInfo.prompt
    #     self.filelogger.info(f'prompt: {prompt}')
    #     res_tag = self.base_predict(prompt)
    #     prompt_tokens = 0
    #     for prom in prompt:
    #         prompt_tokens += self._get_tokens(request_id, prom)
       
    #     if inferenceInfo == RequestMode.STREAM:
    #         await asyncio.ensure_future(self._process_stream_infence(user_tag, request_id, prompt_tokens, res_tag))
    #         return 

    async def infer(self, tasks: list):
        task_data = {
            "request_ids": {},
            "request_infos": {},
            "params": {},
            "patch_ids": {},
            "user_tags": {},
            "prompts": []
        }
        for idx, inferenceInfo in enumerate(tasks):
            task_data["request_ids"][idx] = inferenceInfo.request_id
            task_data["request_infos"][idx] = inferenceInfo.requestInfo
            task_data["params"][idx] = task_data["request_infos"][idx].params
            task_data["user_tags"][idx] = task_data["request_infos"][idx].user_tag
            task_data["patch_ids"][idx] = str(task_data["params"][idx].get("patch_id", 0))
            task_data["prompts"].extend(inferenceInfo.prompt)  
        
        self.filelogger.info(f'prompts: {task_data["prompts"]}')
        res_tags = await self.base_predict(task_data["prompts"])
        tokens = []
        for i, prom in enumerate(task_data["prompts"]):
            tokens.append(self._get_tokens(task_data["request_ids"], prom))

        self.tag_return(res_tags, task_data["user_tags"], tokens)
        
        # for i, res_tag in enumerate(res_tags):
        #     inferenceInfo = tasks[i]
        #     task_tag = [str(res_tag)]
        #     self.filelogger.info(f'mode: {inferenceInfo.mode}')
        #     if inferenceInfo.mode == RequestMode.STREAM:
        #         await asyncio.ensure_future(self._process_stream_infence(user_tag[i], request_id[i], prompt_tokens[i], task_tag))
        #     return 

    def tag_return(self, res_list: list, user_tags, tokens):
        for i, res_tag in enumerate(res_list):
            res = Response()
            resData = ResponseData()
            resData.setData(json.dumps(res_tag, ensure_ascii=False))
            resData.setDataType(DataText)
            resData.status = DataOnce
            res.list = [resData]
            self.filelogger.info(f'user_tag: {user_tags[i]}, res: {res.list[0].data}, prompt_tokens: {tokens[i]}')

    @time_cost
    async def base_predict(self, req_context: list):
        predictions = self.model.predict_scores(req_context)
        # self.filelogger.info(f'predictions: {predictions}')
        res_list = []
        for i, p in enumerate(predictions):
            self.filelogger.info(f'text_i: {i}')
            category_scores = {}
            for idx, prob in enumerate(p):
                # self.filelogger.info(f'idx: {idx}, prob: {prob}')
                category = self.tag_data.get(str(idx), None)
                if category:
                    category_scores[category] = float(prob)
            res_tag = {
                "category_scores": category_scores
            }
            res_list.append(res_tag)
        self.filelogger.info(f'res_list: {res_list}')
        return res_list

    def _get_tokens(self, request_id, prompt):
        prompt_tokens = self.tokenizer.encode(prompt, add_special_tokens=False)
        return len(prompt_tokens)

    async def _process_stream_infence(self, user_tag, request_id, prompt_tokens_len, res_tag):
        self.filelogger.info(f'streaming inference start output: {request_id}') 
        state = DataNone
        last_result = ""
        current_result = res_tag
        index = 0
        try:
            res = Response()
            if state == DataNone:
                state = DataBegin
            elif state == DataBegin:
                state = DataContinue

            resData = ResponseData()
            resData.setData(json.dumps(res_tag, ensure_ascii=False))
            resData.setDataType(DataText)
            resData.status = state
            res.list = [resData]
            last_result = current_result
            index += 1
        except:
            import traceback
            traceback.print_exc()
            self.filelogger.error(f'streaming inference exception, {request_id}')
            exit(-1) 
        result_tokens_len = 0

        # 结束
        state = DataEnd
        content = resp_content(state, index, ' ')
        result_tokens_len = self._get_tokens(request_id, last_result)
        usage = resp_usage(state, prompt_tokens_len, result_tokens_len)
        res = Response()
        res.list = [content, usage]
        self.filelogger.info(
            f'====> streaming inference end, {request_id}:[{current_result}], state:{state}, in_tokens:{prompt_tokens_len}, out_tokens:{result_tokens_len}')